;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;    alist-defthms.lisp
;;;
;;;    Bill Bevier
;;;    Computational Logic, Inc.
;;;    1717 West 6th Street, Suite 290
;;;    Austin, Texas 78703
;;;    bevier@cli.com
;;;
;;;    Copyright 1994, Computational Logic, Inc. All Rights Reserved
;;;    
;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(in-package "ACL2")
;; 
(include-book "alist-defuns")
(include-book "list-defthms")
(include-book "sets")
(local (include-book (:relative "arithmetic" "equalities")))

(deflabel alist-defthms-section)

; The most important lemmas in the section explain the
; meaning of BOUND? and BINDING for each way to construct
; an alist. 

; ---------- MEMBER-EQUAL ----------

(defthm member-equal-domain
  (implies (alistp a)
	   (iff (member-equal x (domain a))
		(bound? x a)))
  :hints (("Goal" :induct (assoc-equal x a)
	   :in-theory (enable domain bound?))))

(defthm member-equal-binding-range
  (implies (bound? x a)
	   (member-equal (binding x a)
			 (range a)))
  :hints (("Goal" :induct (assoc-equal x a)
	   :in-theory (enable binding range bound?))))

; ---------- BOUND? ----------

(local (defun double-list-induction (a b)
	 (declare (xargs :guard (and (true-listp a) (true-listp b))))
	 (cond ((endp a) nil)
	       ((endp b) nil)
	       (t (double-list-induction (cdr a) (cdr b))))))

(defthm equal-domain-fwd-to-bound?
  (implies (and (equal (domain a) (domain b))
		(bound? x a)
		(alistp a)
		(alistp b))
	   (bound? x b))
  :rule-classes :forward-chaining
  :hints (("Goal" :induct (double-list-induction a b)
	   :in-theory (enable bound? domain))))

(defthm bound?-bind 
  (equal (bound? x (bind y v a))
	 (or (equal x y)
	     (bound? x a)))
  :hints (("Goal" :in-theory (enable bound?))))

(defthm consp-assoc-equal-bind
  (equal (consp (assoc-equal x (bind y v a)))
	 (or (equal x y)
	     (consp (assoc-equal x a))))
  :hints (("Goal" :do-not-induct t
	   :use (:instance bound?-bind)
	   :in-theory (union-theories '(bound?)
				      (set-difference-theories
				       (current-theory :here)
				       '(bound?-bind))))))

(defthm bound?-fbind 
  (equal (bound? x (fbind y v a))
	 (or (equal x y)
	     (bound? x a)))
  :hints (("Goal" :in-theory (enable bound? fbind))))

(defthm bound?-rembind
  (equal (bound? x (rembind y a))
	 (and (not (equal x y))
	      (bound? x a)))
  :hints (("Goal" :in-theory (enable bound?))))

(defthm bound?-pairlis$
  (iff (bound? x (pairlis$ a b))
       (member-equal x a))
  :hints (("Goal" :in-theory (enable bound?))))

(defthm bound?-domain-restrict
  (equal (bound? x (domain-restrict l a))
	 (and (member-equal x l)
	      (bound? x a)))
  :hints (("Goal" :induct t :in-theory (enable bound?))))

(defthm bound?-domain-anti-restrict
  (equal (bound? x (domain-anti-restrict l a))
	 (and (not (member-equal x l))
	      (bound? x a)))
  :hints (("Goal" :induct t :in-theory (enable bound?))))

;; BOUND?-ALIST-COMPOSE-WRT-DOMAIN-HELP is new in v1.8.  I am not sure
;; how BOUND?-ALIST-COMPOSE-WRT-DOMAIN was proved without a helper lemma.
;; [Actually I am, it took three inductions in the proof.]
;; A. Flatau  8-Nov-1994
(local
 (defthm bound?-alist-compose-wrt-domain-help
   (implies (or (not (consp (assoc-equal (cdr (assoc-equal x a1)) a2)))
		(not (consp (assoc-equal x a1))))
	    (not (consp (assoc-equal x
				     (alist-compose-wrt-domain dom a1 a2)))))
   :hints (("Goal" :in-theory (enable binding bound?)))))

(defthm bound?-alist-compose-wrt-domain
  (iff (bound? x (alist-compose-wrt-domain dom a1 a2))
       (and (member-equal x dom)
	    (bound? x a1)
	    (bound? (binding x a1) a2)))
  :hints (("Goal" :induct t :in-theory (enable bound? binding))))

(defthm bound?-alist-invert
  (iff (bound? x (alist-invert a))
       (member-equal x (range a)))
  :hints (("Goal" :induct t :in-theory (enable bound? range))))

; ---------- DOMAIN ----------

(defthm domain-bind
  (implies (alistp a)
	   (equal (domain (bind x v a))
		  (if (bound? x a)
		      (domain a)
		      (append (domain a) (list x)))))
  :hints (("Goal" :induct t
	   :in-theory (enable bound? domain))))

(defthm domain-fbind 
  (equal (domain (fbind x v a))
	 (cons x (domain a)))
  :hints (("Goal" :do-not-induct t 
	   :in-theory (enable fbind bound? domain))))

(defthm domain-rembind
  (equal (domain (rembind x a))
	 (remove-equal x (domain a)))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain))))

(defthm domain-pairlis$
  (implies (true-listp a)
	   (equal (domain (pairlis$ a b)) a))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain))))

(defthm domain-alist-invert 
  (equal (domain (alist-invert a)) (range a))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain range))))

(defthm domain-domain-restrict
  (equal (domain (domain-restrict l a))
	 (intersection-equal (domain a) l))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain))))

(defthm domain-domain-anti-restrict
  (equal (domain (domain-anti-restrict l a))
	 (set-difference-equal (domain a) l))
  :hints (("Goal" :induct t
	   :in-theory (enable domain))))

; ---------- BINDING ----------

(defthm binding-bind
  (equal (binding x (bind y v a))
	 (if (equal x y)
	     v
	     (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding))))

(defthm binding-fbind
  (equal (binding x (fbind y v a))
	 (if (equal x y)
	     v
	     (binding x a)))
  :hints (("Goal" :do-not-induct t
	   :in-theory (enable binding fbind))))

(defthm binding-rembind
  (equal (binding x (rembind y a))
	 (if (equal x y)
	     nil
	     (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding))))

(defthm binding-pairlis$
  (equal (binding x (pairlis$ a b))
	 (if (member-equal x a)
	     (nth (position$ x a) b)
	     nil))
  :hints (("Goal" :induct t
	   :in-theory (enable binding))))

(defthm binding-domain-restrict 
  (implies (and (member-equal x l)
		(bound? x a))
	   (equal (binding x (domain-restrict l a))
		  (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding bound?))))

(defthm binding-domain-anti-restrict 
  (implies (and (not (member-equal x l))
		(bound? x a))
	   (equal (binding x (domain-anti-restrict l a))
		  (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding bound?))))

(defthm binding-alist-compose-wrt-domain
  (implies (and (member-equal x dom)
		(bound? x a1)
		(bound? (binding x a1) a2))
	   (equal (binding x (alist-compose-wrt-domain dom a1 a2))
		  (binding (binding x a1) a2)))
  :hints (("Goal" :induct (alist-compose-wrt-domain dom a1 a2)
	   :in-theory (enable bound?))))

(defthm binding-alist-invert
  (implies (member-equal x (range a))
	   (equal (binding x (alist-invert a))
		  (nth (position$ x (range a)) (domain a))))
  :hints (("Goal" :induct t
	   :in-theory (enable binding domain range))))

; ---------- RANGE ----------

(DEFTHM RANGE-BIND
  (IMPLIES (and (ALISTP A)
		(not (BOUND? X A)))
	   (EQUAL (RANGE (BIND X V A))
		  (APPEND (RANGE A) (LIST V))))
  :HINTS
  (("Goal" :INDUCT T :IN-THEORY
    (ENABLE BOUND? DOMAIN
	    RANGE POSITION BIND ALISTP PUT-NTH))))


; ---------- SUBSETP-EQUAL ----------

(defthm subsetp-equal-all-bindings-range
  (subsetp-equal (all-bindings l a) (range a)))

; ---------- ALL-BOUND? ----------

(defthm all-bound?-bind
  (implies (all-bound? l a)
	   (all-bound? l (bind x y a))))

(defthm all-bound?-append
  (equal (all-bound? (append x y) a)
	 (and (all-bound? x a)
	      (all-bound? y a))))

(defthm all-bound?-firstn
  (implies (all-bound? l a)
	   (all-bound? (firstn n l) a)))

(defthm all-bound?-nthcdr
  (implies (all-bound? l a)
	   (all-bound? (nthcdr n l) a)))

; ---------- MAKE-ALIST & FLATTEN-ALIST ----------

(defthm make-alist-flatten-alist
  (implies (alistp a)
	   (equal (make-alist (flatten-alist a))
		  a)))



; ========================================

(deftheory alist-defthms
  (set-difference-theories (current-theory :here)
			   (current-theory 'alist-defthms-section)))

