;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;    sets.lisp
;;;
;;;    Bill Bevier
;;;    Computational Logic, Inc.
;;;    1717 West 6th Street, Suite 290
;;;    Austin, Texas 78703
;;;    bevier@cli.com
;;;
;;;    Copyright 1994, Computational Logic, Inc. All Rights Reserved
;;;    
;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(in-package "ACL2")

; ------------------------------------------------------------
; DEFINITIONS
; ------------------------------------------------------------

; intersection-equal (a b)   intersection operation
; member-equal (x l)         set membership operation
; set-difference-equal (a b) set difference operation
; set-equal (a b)            set equivalence relation  
; subsetp-equal (a b)        subset relation on two lists
; union-equal (a b)          union operation
; excess-element (a b)       find an element in a that's not in b

(defun intersection-equal (x y)
  (declare (xargs :guard (and (true-listp x) (true-listp y))))
  (cond ((endp x) nil)
	((member-equal (car x) y)
	 (cons (car x) (intersection-equal (cdr x) y)))
	(t (intersection-equal (cdr x) y))))

; Note: Set-equal cannot be declared as an equivalence relation
; in Acl2 because it has guards.

(defun set-equal (a b)
  (declare (xargs :guard (and (true-listp a)
			      (true-listp b))))
  (and (subsetp-equal a b)
       (subsetp-equal b a)))

(defun excess-element (s1 s2)
  (declare (xargs :guard (and (true-listp s1)
			      (true-listp s2))))
  (cond ((endp s1) nil)
	((member-equal (car s1) s2)
	 (excess-element (cdr s1) s2))
	(t (car s1))))

; ------------------------------------------------------------
; Theorems
; ------------------------------------------------------------

; Equivalence relation properties of Set-equal (conditional on
; guards).

(defthm subsetp-equal-cons 
  (implies (subsetp-equal a b)
	   (subsetp-equal a (cons x b))))

(defthm subsetp-equal-reflexive 
  (subsetp-equal l l))

(defthm subsetp-equal-transitive 
  (implies (and (subsetp-equal a b)
		(subsetp-equal b c))
	   (subsetp-equal a c))
  :rule-classes nil)

(defthm set-equal-reflexive
  (set-equal l l))

(defthm set-equal-symmetric 
  (implies (set-equal a b)
	   (set-equal b a))
  :rule-classes nil)

(defthm set-equal-transitive 
  (implies (and (set-equal a b)
		(set-equal b c))
	   (set-equal a c))
  :rule-classes nil
  :hints (("Goal" :do-not-induct t
	   :use ((:instance subsetp-equal-transitive (a a) (b b) (c c))
		 (:instance subsetp-equal-transitive (a c) (b b) (c a))))))

; ---------- MEMBER-EQUAL ----------

(defthm member-equal-subsetp-equal 
  (implies (and (member-equal x a)
		(subsetp-equal a b))
	   (member-equal x b))
  :rule-classes ())

(defthm member-equal-set-equal
  (implies (set-equal a b)
	   (iff (member-equal x a)
		(member-equal x b)))
  :rule-classes ()
  :hints (("Goal" :do-not-induct t
	   :use ((:instance member-equal-subsetp-equal (x x) (a a) (b b))
		 (:instance member-equal-subsetp-equal (x x) (a b) (b a))))))

(defthm member-equal-intersection-equal
  (iff (member-equal x (intersection-equal a b))
       (and (member-equal x a)
	    (member-equal x b))))

(defthm member-equal-set-difference-equal
  (iff (member-equal x (set-difference-equal a b))
       (and (member-equal x a)
	    (not (member-equal x b)))))

(defthm member-equal-union-equal
  (iff (member-equal x (union-equal a b))
       (or (member-equal x a)
	   (member-equal x b))))

(defthm member-equal-excess-element
  (implies (not (subsetp-equal s1 s2))
	   (not (member-equal (excess-element s1 s2) s2))))