
(in-package "ACL2")
(deflabel alist-defuns-section)

(defun alist-invert (a)
  "The alist whose domain is A's range, and whose range is A's domain."
  (declare (xargs :guard (alistp a)))
  (cond ((endp a) nil)
	(t (cons (cons (cdr (car a)) (car (car a)))
		 (alist-invert (cdr a))))))

(defun bind (x y a)
  "The alist derived from A by binding X to Y."
  (declare (xargs :guard (alistp a)))
  (cond ((endp a) (list (cons x y)))
	((equal x (car (car a))) (cons (cons x y) (cdr a)))
	(t (cons (car a) (bind x y (cdr a))))))

(defun binding (x a)
  "The value bound to X in alist A."
  (declare (xargs :guard (alistp a)))
  (cdr (assoc-equal x a)))

(defun boundp (x a)
  "Is X bound in alist A?"
  (declare (xargs :guard (alistp a)))
  (consp (assoc-equal x a)))

(defun domain (a)
  "The list of CARs of an alist."
  (declare (xargs :guard (alistp a)))
  (strip-cars a))

(defun domain-anti-restrict (l a)
  "An alist containing only those pairs in A whose CAR is not in L."
  (declare (xargs :guard (and (true-listp l)
			      (alistp a))))
  (cond ((endp a) nil)
	((member-equal (car (car a)) l)
	 (domain-anti-restrict l (cdr a)))
	(t (cons (car a) (domain-anti-restrict l (cdr a))))))

(defun domain-restrict (l a)
  "An alist containing only those pairs in A whose CAR is in L."
  (declare (xargs :guard (and (true-listp l)
			      (alistp a))))
  (cond ((endp a) nil)
	((member-equal (car (car a)) l)
	 (cons (car a) (domain-restrict l (cdr a))))
	(t (domain-restrict l (cdr a)))))

(defun fbind (x v a)
  "Bind key X to value V in alist A (fast)."
  (declare (xargs :guard (alistp a)))
  (cons (cons x v) a))

(defun flatten-alist (a)
  (declare (xargs :guard (alistp a)
		  :verify-guards nil))
  (cond ((endp a) nil)
	(t (cons (caar a) (cons (cdar a) (flatten-alist (cdr a)))))))

(defun make-alist (l)
  (declare (xargs :guard (true-listp l)
		  :verify-guards nil))
  (cond ((endp l) nil)
	((endp (cdr l)) nil)
	(t (cons (cons (car l) (cadr l))
		 (make-alist (cddr l))))))

(defun range (a)
  "The list of CDRs of an alist."
  (declare (xargs :guard (alistp a)))
  (strip-cdrs a))

(defun rembind (x a)
  "The alist derived from A by removing any binding of X."
  (declare (xargs :guard (alistp a)))
  (cond ((endp a) nil)
	((equal x (car (car a))) (rembind x (cdr a)))
	(t (cons (car a) (rembind x (cdr a))))))

(defun all-bindings (l a)
  "The list of bindings of elements of list L in alist A."
  (declare (xargs :guard (and (true-listp l)
			      (alistp a))))  
  (cond ((endp l) nil)
	((bound? (car l) a) 
	 (cons (binding (car l) a) (all-bindings (cdr l) a)))
	(t (all-bindings (cdr l) a))))

(defun all-boundp (l a)
  "Are all elements of list L bound in alist A?"
  (declare (xargs :guard (and (true-listp l)
			      (alistp a))))
  (cond ((endp l) t)
	(t (and (bound? (car l) a)
		(all-bound? (cdr l) a)))))

(defun alist-compose-wrt-domain (dom a1 a2)
  "X is bound to Z in (ALIST-COMPOSE DOM A1 A2) if X occurs in
   DOM, X is bound in A1, and there exists a Y such that
   (BINDING X A1) = Y, and (BINDING Y A2) = Z."
  (declare (xargs :guard (and (true-listp dom)
			      (alistp a1)
			      (alistp a2))
		  :verify-guards nil))
  (cond ((endp dom) nil)
	((and (bound? (car dom) a1)
	      (bound? (binding (car dom) a1) a2))
	 (bind (car dom) 
	       (binding (binding (car dom) a1) a2)
	       (alist-compose-wrt-domain (cdr dom) a1 a2)))
	(t (alist-compose-wrt-domain (cdr dom) a1 a2))))

(defmacro alist-compose (a1 a2)
  `(alist-compose-wrt-domain (domain ,a1) ,a1 ,a2))

(defthm alistp-bind
  (implies (alistp a)
	   (alistp (bind x v a)))
  :hints (("Goal" :induct (alistp a))))

(defthm alistp-fbind
  (implies (alistp a)
	   (alistp (fbind x v a)))
  :hints (("Goal" :induct (alistp a))))

(defthm alistp-rembind
  (implies (alistp a)
	   (alistp (rembind x a)))
  :hints (("Goal" :induct (rembind x a))))

(defthm alistp-make-alist
  (alistp (make-alist l))
  :hints (("Goal" :induct (make-alist l))))

(defthm alistp-pairlis 
  (alistp (pairlis a b))
  :hints (("Goal" :induct (pairli a b))))

(defthm alistp-alist-invert 
  (alistp (alist-invert a)))

(defthm alistp-domain-restrict
  (implies (alistp a)
	   (alistp (domain-restrict l a))))
 
(defthm alistp-domain-anti-restrict
  (implies (alistp a)
	   (alistp (domain-anti-restrict l a))))

(defthm alistp-alist-compose-wrt-domain 
  (alistp (alist-compose-wrt-domain dom a1 a2)))

(verify-guards alist-compose-wrt-domain)

(deftheory alist-non-recursive-defuns
  '(binding bound domain range fbind))

(defthm true-listp-flatten-alist
  (true-listp (flatten-alist l)))

(in-theory (disable alist-non-recursive-defuns))
  
(deftheory alist-defuns
  (set-difference-theories (current-theory :here)
			   (union-theories 
			    (current-theory 'alist-defuns-section)
			    (theory 'alist-non-recursive-defuns))))



