; -*-Lisp-*-
; ---------------------------------------------------------------------------
; 
; File:         save-iacl2.lisp
; Description:  Load and save and Infix ACL2 executable.
; Author:       R. Cohen
; Created:      Wed Jan 29 13:04:14 1997
; Modified:     Mar  3 97, (MKSmith)
; Language:     Lisp
; Package:      N/A
; Status:       Experimental (but Distributable)
; 
; (C) Copyright 1997, Computational Logic, Inc., all rights reserved.
; 
; ---------------------------------------------------------------------------

#| To create an IACL2 core image

    acl2 < save-iacl2.lisp

  This will load everything into ACL2, and then save the core image.

  Loading the ACL2 files (i.e., certified books) may take 5 - 10 minutes
  or more depending on your system.
|#

(include-book "infix-prelude")

;; Be sure we're at the Lisp command loop.

:q

;; The system initialization code below only works in GCL.  
;; Complain if it's run in another Lisp.

#-GCL
(error "Saving an Infix image is currently only supported in GCL.
You will need to edit the system dependencies in save-iacl2.lisp
before you can build Iacl2.")

;; This code only works in version of ACL2 >= 1.9

(defvar *minimum-acl2-version-number* 1.89
  "Due to rounding we may get something like 1.899.. when reading 1.9")

(defun later-1.89 (version)
  (let ((number (read-from-string 
		 (subseq version (- (length version) 4) (length version)))))

    (>= number *minimum-acl2-version-number*)))

(if (and (boundp '*acl2-version*) (stringp *acl2-version*))
    (if (later-1.89 *acl2-version*)
	nil
	(error "We don't support versions of ACL2 prior to 1.9"))
    (format *terminal-io* "WARNING: The variable *ACL2-VERSION* is not bound
It is likely that the version you are using is prior to 1.9 and 
will not interact properly with this infix version."))

;; Turn off messages from the garbage collector.

(setq *gbc-notify* nil)
(setq *gbc-message* nil)

(load "infix-interface.lisp")

(defparameter *init-loaded* t)

;; To restart the ACL2 command loop. (ACL2::LP)

;; The init.lsp and acl2-customization files have already loaded
;; everything.

(defparameter iacl2-Version "15.0")

(defparameter iacl2-Sysout-Name 
  (if (fboundp (find-symbol "MI-PATHNAME-PART" "USER"))
      ;; If mi-load.lisp is loaded, then identify the h/w & s/w
      ;; architecture under which this executable was saved.
      (let ((fn (find-symbol "MI-PATHNAME-PART" "USER")))
	(format nil "iacl2.~a" (funcall (find-symbol "MI-PATHNAME-PART" "USER"))))
    ;; If mi-load.lisp isn't loaded, just  use the version name.
    "iacl2")
  "File name for the saved executable file.")

(defparameter iacl2-Sysout-Comment (format nil "iacl2 ~a" iacl2-Version)
  "Comment printed as part of the welcome banner of the saved file.")

(defvar *acl2-allocation-alist* nil)

(defparameter *iacl2-saved-string*
"~% ~a built ~a.~
    ~a~%~% NOTE!!  Proof trees are disabled in ACL2.  To enable them in emacs,~
    ~% look under the ACL2 source directory in interface/emacs/README.doc; ~
    ~% and, to turn on proof trees, execute :START-PROOF-TREE in the ACL2 ~
    ~% command loop.   Look in the ACL2 documentation under PROOF-TREE.~%")

;;
;; Now define the function to save a GCL image.
;; This function will save the IACL2 image with its loaded books.

(defun save-iacl2 (&optional (sysout-name  "iacl2") (mode "Infix Libraries"))
  (setq si::*top-level-hook*
	#'(lambda ()
	    (format t *iacl2-saved-string*
		    *copy-of-acl2-version*
		    *saved-build-date*
		    (cond (mode
			   (format nil "~% Initialized with ~a." mode))
			  (t "")))
	    (let ((fl (probe-file (merge-pathnames (user-homedir-pathname)
						   "iacl2-init.lsp"))))
	      (if fl (progn (terpri) (load fl)) (format t "~% No iacl2-init.lsp file loaded.~%")))
	    (cond
	     (*acl2-allocation-alist*
	      (format t
		      "~% Beginning allocations.
 Set acl2::*acl2-allocation-alist* to NIL in ~~/acl2-init.lsp
 if you must make your running image smaller.~%")
	      (sloop::sloop for (type . n) in *acl2-allocation-alist*
			    when n
			    do
			    ;; (format t "Allocating ~s to ~s.~%" type n)
			    (let ((x (symbol-name type)))
			      (cond
			       ((equal x "HOLE")
				(si::set-hole-size n))
			       ((equal x "RELOCATABLE")
				(si::allocate-relocatable-pages n))
			       (t (si::allocate type n t)))))))
	    (in-package "ACL2")
	    (assign infixp t)
	    (acl2::lp)))
  (si::save-system sysout-name))

(save-iacl2 iacl2-Sysout-Name iacl2-Sysout-Comment)

;;
;; End of File
;;
